/* The FreeDOS-32 FAT Driver version 2.0
 * Copyright (C) 2001-2005  Salvatore ISAJA
 *
 * This file "alloc.c" is part of the FreeDOS-32 FAT Driver (the Program).
 *
 * The Program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * The Program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Program; see the file GPL.txt; if not, write to
 * the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/**
 * \addtogroup fat
 * @{
 */
/**
 * \file
 * \brief Manages allocation of clusters and the File Allocation Table.
 */
#include "fat.h"


/******************************************************************************
 * Seek functions.
 * Find the location of the "n"-th cluster entry in the file allocation table.
 * On success, put the number of the sector containing the entry in "sector"
 * and return the byte offset of the entry in that sector.
 ******************************************************************************/

static int fat12_seek(Volume *v, Cluster n, unsigned fat_num, Sector *sector)
{
	Cluster offset;
	if (n > v->data_clusters + 1) return -ENXIO;
	offset = n + (n >> 1); /* floor(n * 1.5) */
	*sector = fat_num * v->fat_size + v->fat_start + (offset >> v->log_bytes_per_sector);
	return offset & (v->bytes_per_sector - 1); /* % bytes_per_sector */
}


static int fat16_seek(Volume *v, Cluster n, unsigned fat_num, Sector *sector)
{
	Cluster offset;
	if (n > v->data_clusters + 1) return -ENXIO;
	offset = n << 1;
	*sector = fat_num * v->fat_size + v->fat_start + (offset >> v->log_bytes_per_sector);
	return offset & (v->bytes_per_sector - 1); /* % bytes_per_sector */
}


static int fat32_seek(Volume *v, Cluster n, unsigned fat_num, Sector *sector)
{
	Cluster offset;
	if (n > v->data_clusters + 1) return -ENXIO;
	offset = n << 2;
	*sector = fat_num * v->fat_size + v->fat_start + (offset >> v->log_bytes_per_sector);
	return offset & (v->bytes_per_sector - 1); /* % bytes_per_sector */
}


/******************************************************************************
 * Read functions.
 * Read the value of the specified cluster entry of the file allocation table.
 * On success, return the not negative entry value.
 ******************************************************************************/


/**
 * \brief  Reads a cluster entry from the active copy of the FAT.
 * \param  v the volume the FAT to read belongs to;
 * \param  n number of the cluster entry to read.
 * \return The non-negative value of the cluster entry, or a negative error.
 */
static int32_t read_fat_entry(Volume *v, Cluster n)
{
	return v->fat_read(v, n, v->active_fat);
}


int32_t fat12_read(Volume *v, Cluster n, unsigned fat_num)
{
	int     res, offset;
	Sector  sector;
	Cluster c;
	Buffer *b = NULL;

	offset = fat12_seek(v, n, fat_num, &sector);
	if (offset < 0) return offset;
	res = fat_readbuf(v, sector, &b, false);
	if (res < 0) return res;
	if (offset < v->bytes_per_sector - 1) {
		//c = (Cluster) *((uint16_t *) &b->data[offset + res]);
        c = (Cluster) b->data[offset + res] + ((Cluster) b->data[offset + res+1] << 8) ;
        if (c==0) {
            c = (Cluster) *((uint16_t *) &b->data[offset + res]);
         //    if (c==0)
           //     c =  *((uint16_t *) &b->data[offset + res]);
            //c = (Cluster) b->data[offset + res];
            printf("->-> zero 1, c=%d, offset=%d, res=%d\n", c, offset, res);
        }
        
    }
	else /* the entry spans across two sectors */
	{
		Buffer *b2 = NULL;
		int res2 = fat_readbuf(v, sector + 1, &b2, false);
		if (res2 < 0) return res2;
		c = (Cluster) b->data[offset + res] + ((Cluster) b2->data[res2] << 8);
           if (c==0) {
                 printf("->-> zero 1, c=%d, offset=%d, res=%d\n", c, offset, res);
        }
        
                
	}
	/* If the entry number is odd we need the highest 12 bits of "c",
	 * whereas if it's even we need the lowest 12 bits. */
	if (n & 1) c >>= 4; else c &= 0x0FFF;
       // printf("Value of  c =%d", c);
	if (IS_FAT12_EOC(c)) c = FAT_EOC;
   // printf("Value of  c =%d", c);
    
	return c;
}


int32_t fat16_read(Volume *v, Cluster n, unsigned fat_num)
{
	int     res, offset;
	Sector  sector;
	Cluster c;
	Buffer *b = NULL;

	offset = fat16_seek(v, n, fat_num, &sector);
	if (offset < 0) return offset;
	res = fat_readbuf(v, sector, &b, false);
	if (res < 0) return res;
	c = (Cluster) *((uint16_t *) &b->data[offset + res]);
	if (IS_FAT16_EOC(c)) c = FAT_EOC;
	return c;
}


int32_t fat32_read(Volume *v, Cluster n, unsigned fat_num)
{
	int     res, offset;
	Sector  sector;
	Cluster c;
	Buffer *b = NULL;

	offset = fat32_seek(v, n, fat_num, &sector);
	if (offset < 0) return offset;
	res = fat_readbuf(v, sector, &b, false);
	if (res < 0) return res;
	c = (Cluster) *((uint32_t *) &b->data[offset + res]) & 0x0FFFFFFF;
	if (IS_FAT32_EOC(c)) c = FAT_EOC;
	return c;
}


/**
 * \brief   Gets the address of a sector of a file.
 * \param   c            the channel to get the sector of;
 * \param   sector_index index of the sector of the file to get the address of;
 * \param   sector       to receive the address of the sector corresponding to \c sector_index;
 * \return  0 on success, >0 on EOF, or a negative error.
 * \remarks \c c->cluster_index and \c c->cluster are updated to cache the address of the last
 *          cluster of the file reached by this function. On EOF, they relate to the last
 *          cluster of the file, if any.
 */
int fat_get_file_sector(Channel *c, Sector sector_index, Sector *sector)
{
	int res = 0;
	File   *f = c->f;
	Volume *v = f->v;
	if (!f->de_sector && !f->first_cluster) /* FAT12/FAT16 root directory */
	{
		if (sector_index >= v->root_size) return 1;
		*sector = sector_index + v->root_sector;
	}
	else /* File/directory with linked allocation */
	{
		Cluster  cluster_index     = sector_index >> v->log_sectors_per_cluster;
		unsigned sector_in_cluster = sector_index & (v->sectors_per_cluster - 1);
		if (!f->first_cluster) return 1;
		if (!c->cluster_index || (cluster_index < c->cluster_index))
		{
			c->cluster_index = 0;
			c->cluster = f->first_cluster;
		}
		while (c->cluster_index < cluster_index)
		{
			int32_t nc = read_fat_entry(v, c->cluster);
			if (nc < 0) {
                printf("2. NC= %d\n", nc); 
                return nc;
            }
			if (nc == FAT_EOC) return 1;
			if ((nc < 2) || (nc > v->data_clusters + 1)) {
                printf("3. nc = %d, v->data_clusters = %d\n", nc, v->data_clusters);
                return -EIO;
            }
			c->cluster_index++;
			c->cluster = (Cluster) nc;
		}
		*sector = ((c->cluster - 2) << v->log_sectors_per_cluster)
		        + v->data_start + sector_in_cluster;
	}
	return res;
}

/* @} */
